InputData = $00
OutputData = $02
PastData = $04
FlagBuffer = $06
FlagCounter = $07
MatchLength = $08

	.org $bff0
	.db "NES", $1a	;id
	.db $01		;16 PRG-ROM pages
	.db $00		;no CHR-ROM present
	.db $21		;mapper 2, vertical mirroring
	.dcb $09, $00	;clear the rest

	.org $c000
Reset:
	sei			;Disable interrupts
	ldx #$00
	stx $2000		;Reset the PPU
	stx $2001
	cld			;Make sure we're not in decimal mode
	dex
	txs			;Initialize the stack

WaitVBlank1:			;Wait for the PPU to become stable
	bit $2002
	bpl WaitVBlank1
WaitVBlank2:			;Wait one more frame, to be safe
	bit $2002
	bpl WaitVBlank2

	lda #<GFXBlock
	sta InputData+0
	lda #>GFXBlock
	sta InputData+1
	lda #$00
	sta OutputData+0
	lda #$00
	sta OutputData+1

	jsr Decompress

	lda #$3f
	sta $2006
	lda #$00
	sta $2006

	lda #$01
	sta $2007
	lda #$11
	sta $2007
	lda #$21
	sta $2007
	lda #$31
	sta $2007

	lda #$08
	sta $2001

Forever:
	jmp Forever

	;InputData is a pointer to the compressed data
	;OutputData is the destination address
Decompress:
	lda OutputData+1
	sta $2006
	lda OutputData+0
	sta $2006
	ldy #$00
.NewFlagBuffer:
	lda #$08
	sta FlagCounter
	jsr .ReadByte
	sta FlagBuffer
.DecodeBlock:
	asl FlagBuffer
	bcs .CompressedBlock
.UncompressedBlock:
	jsr .ReadByte
	jsr .WriteByte
.NextBlock:
	dec FlagCounter
	bne .DecodeBlock
	beq .NewFlagBuffer
.CompressedBlock:
	lda (InputData), y
	bpl .LongMatch
.ShortMatch:
	and #$60
	lsr
	lsr
	lsr
	lsr
	lsr
	adc #$02
	sta MatchLength
	jsr .ReadByte
	and #$1f
	eor #$ff
	adc OutputData+0
	sta PastData+0
	lda #$ff
	adc OutputData+1
	sta PastData+1
	bcs .CopyData
.LongMatch:
	beq .Finished
	and #$78
	lsr
	lsr
	lsr
	adc #$01
	sta MatchLength
	jsr .ReadByte
	and #$07
	eor #$ff
	pha
	jsr .ReadByte
	eor #$ff
	adc OutputData+0
	sta PastData+0
	pla
	adc OutputData+1
	sta PastData+1
.CopyData:
	clc
	lda PastData+1
	sta $2006
	lda PastData+0
	sta $2006
	adc #$01
	sta PastData+0
	lda PastData+1
	adc #$00
	sta PastData+1
	lda $2007
	lda $2007
	pha
	lda OutputData+1
	sta $2006
	lda OutputData+0
	sta $2006
	pla
	jsr .WriteByte
	dec MatchLength
	bne .CopyData

	jmp .NextBlock
;	dec FlagCounter
;	bne .DecodeBlock
;	beq .NewFlagBuffer
.Finished:
	rts
.ReadByte:
	lda (InputData), y
	iny
	bne .NoWrapR
	inc InputData+1
.NoWrapR:
	rts
.WriteByte:
	sta $2007
	inc OutputData+0
	bne .NoWrapW
	inc OutputData+1
.NoWrapW:
	rts

NMI:
	rti
IRQ:
	rti

GFXBlock:
	.include "smbchr.asm"

	;Interrupt vectors
	.org $fffa
	.dw NMI
	.dw Reset
	.dw IRQ