Option Explicit

Class ByteStream
	Private BinaryString
	Private BinaryPosition
	Private BinarySize

	Private Sub Class_Initialize
		BinaryPosition = 1
		BinarySize = 0
		BinaryString = ""
	End Sub

	Private Sub Class_Terminate
		BinaryString = ""
	End Sub

	Public Sub Write(StringToWrite)
		If BinaryPosition = BinarySize + 1 Then 'If it's the end of the string, just append the new string
			BinaryString = BinaryString & StringToWrite
			BinarySize = BinarySize + LenB(StringToWrite)
		Else
			Dim RightSize
			RightSize = BinarySize - BinaryPosition - (LenB(StringToWrite) - 1)
			If RightSize < 0 Then
				BinarySize = BinarySize + Abs(RightSize)
				RightSize = 0
			End If
			BinaryString = LeftB(BinaryString, BinaryPosition - 1) & StringToWrite & RightB(BinaryString, RightSize)
		End If
		BinaryPosition = BinaryPosition + LenB(StringToWrite)
	End Sub

	Public Function Read(ByteCount)
		Read = MidB(BinaryString, BinaryPosition, ByteCount)
		BinaryPosition = BinaryPosition + ByteCount
	End Function

	Public Function ReadFrom(PositionToReadFrom, ByteCount)
		ReadFrom = MidB(BinaryString, PositionToReadFrom + 1, ByteCount)
	End Function

	Public Sub WriteByte(ByteToWrite)
		If ByteToWrite < 256 Then Write ChrB(ByteToWrite)
	End Sub

	Public Sub WriteByteAt(PositionToWriteTo, ByteToWrite)
		Dim CurrentPosition
		CurrentPosition = BinaryPosition
		BinaryPosition = PositionToWriteTo + 1
		WriteByte ByteToWrite
		BinaryPosition = CurrentPosition
	End Sub

	Public Function ReadByte
		ReadByte = AscB(Read(1))
	End Function

	Public Function ReadByteAt(PositionToReadFrom)
		ReadByteAt = AscB(ReadFrom(PositionToReadFrom, 1))
	End Function

	Public Sub LoadFromFile(FilePath)
		Dim BinaryStream

		Set BinaryStream = CreateObject("ADODB.Stream")

		With BinaryStream
			.Type = 1 'adTypeBinary = 1
			.Mode = 3 'adModeReadWrite = 3
			.Open
			.LoadFromFile FilePath
			.Position = 0
		End With

		BinaryPosition = 1
		BinarySize = BinaryStream.Size
		BinaryString = LeftB(BinaryStream.Read(-1), BinarySize) 'adReadAll = -1

		Set BinaryStream = Nothing
	End Sub

	Public Sub SaveToFile(FilePath)
		Dim TextStream, BinaryStream

		Set TextStream = CreateObject("ADODB.Stream")
		Set BinaryStream = CreateObject("ADODB.Stream")

		With TextStream
			.Type = 2 'adTypeText = 2
			.CharSet = "Unicode"
			.Mode = 3 'adModeReadWrite = 3
			.Open
			.Position = 0
			.WriteText BinaryString
			.Position = 0
			.Type = 1 'adTypeBinary
			.Position = 2
		End With

		With BinaryStream
			.Type = 1 'adTypeBinary
			.Mode = 3 'adModeReadWrite = 3
			.Open
			.Position = 0
			.Write TextStream.Read(-1) 'adReadAll = -1
			.SaveToFile FilePath, 2 'adSaveCreateNotExist = 1, adSaveCreateOverWrite = 2
		End With

		Set TextStream = Nothing
		Set BinaryStream = Nothing
	End Sub

	Public Sub SaveToAsmFile(FilePath)
		Dim TextStream

		Set TextStream = CreateObject("ADODB.Stream")

		With TextStream
			.Type = 2 'adTypeText = 2
			.CharSet = "us-ascii"
			.Mode = 3 'adModeReadWrite = 3
			.Open
			.Position = 0
		End With

		BinaryPosition = 1

		While BinaryPosition <= BinarySize
			If ((BinaryPosition - 1) Mod 16) = 0 Then
				TextStream.WriteText Chr(9) & ".db "
			Else
				TextStream.WriteText ", "
			End If

			TextStream.WriteText "$" & Right("0" & Hex(ReadByte), 2)

			If ((BinaryPosition - 1) Mod 16) = 0 Then TextStream.WriteText Chr(13) & Chr(10)
		Wend

		TextStream.SaveToFile FilePath, 2 'adSaveCreateNotExist = 1, adSaveCreateOverWrite = 2

		Set TextStream = Nothing
	End Sub

	Property Get Size
		Size = BinarySize
	End Property

	Property Get EndOfData
		If BinaryPosition = BinarySize + 1 Then
			EndOfData = True
		Else
			EndOfData = False
		End If
	End Property

	Property Get Position
		Position = BinaryPosition - 1
	End Property

	Property Let Position(PositionToSet)
		If PositionToSet < BinarySize Then
			BinaryPosition = PositionToSet + 1
		Else
			'Error
		End If
	End Property
End Class

Class BitStream
	Private DataStream
	Private BitBuffer
	Private BitMask

	Private Sub Class_Initialize
		Set DataStream = New ByteStream
		BitMask = Array(&H0001, &H0002, &H0004, &H0008, &H0010, &H0020, &H0040, &H0080, &H0100, &H0200, &H0400, &H0800, &H1000, &H2000, &H4000, &H8000)
		BitBuffer = 256
	End Sub

	Private Sub Class_Terminate
		Set DataStream = Nothing
	End Sub

	Public Sub WriteBit(BitToWrite)
		BitBuffer = (BitBuffer * 2) Or (BitToWrite And 1)
		If (BitBuffer And 65536) > 0 Then
			DataStream.WriteByte BitBuffer And 255
			BitBuffer = 256
		End If
	End Sub

	Public Sub WriteBits(BitsToWrite, ByVal BitCount)
		While BitCount > 0
			WriteBit (BitsToWrite And BitMask(BitCount - 1)) \ BitMask(BitCount - 1)
			BitCount = BitCount - 1
		Wend
	End Sub

	Public Function ReadBit
		If (BitBuffer And 65536) > 0 Then
			If Not(DataStream.EndOfData) Then
				BitBuffer = DataStream.ReadByte Or 256
			Else
				'Error
			End If
		End If
		ReadBit = (BitBuffer And 128) \ 128
		BitBuffer = BitBuffer * 2
	End Function

	Public Function ReadBits(ByVal BitCount)
		Dim ReturnValue
		ReturnValue = 0
		While BitCount > 0
			ReturnValue = (ReturnValue * 2) Or ReadBit
			BitCount = BitCount - 1
		Wend
		ReadBits = ReturnValue
	End Function

	Public Sub LoadFromFile(FilePath)
		DataStream.LoadFromFile FilePath
		BitBuffer = 65536
	End Sub

	Public Sub SaveToFile(FilePath)
		While BitBuffer > 256
			WriteBit 0
		Wend
		DataStream.SaveToFile FilePath
	End Sub
End Class